/*
 * libformat - protection against format string attacks
 *
 * Copyright (c) 2000 Tim Robbins
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */

/*
 * $Log: libformat.c,v $
 * Revision 1.7  2000/11/13 07:17:27  fyre
 * Log only the first 50 characters of the format string
 *
 * Revision 1.6  2000/10/23 10:29:41  fyre
 * Promote to stable state
 * Move copyright notices around
 * libformat: prefix to messages is done only once
 *
 * Revision 1.5  2000/10/06 07:25:31  fyre
 * Remove parentheses from function definitions
 * Wrap err and warn family of functions (BSD)
 *
 * Revision 1.4  2000/10/06 07:09:53  fyre
 * Log process id with fatal errors
 * Use glibc's builtin format string parsing function
 *
 * Revision 1.3  2000/09/20 09:24:24  fyre
 * Fixd Speeling
 *
 * Revision 1.2  2000/09/20 09:20:40  fyre
 * Check for behaviour not defined by C standard
 * Fix infinite recursion problem with fatal()
 *
 */

#include <dlfcn.h>
#include <errno.h>
#include <printf.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>

static char *rcsid = "$Id: libformat.c,v 1.7 2000/11/13 07:17:27 fyre Stab $";

/*
 * Prototypes of libc functions.
 */
typedef int (*vprintf_t) (const char *format, va_list ap);
typedef int (*vsprintf_t) (char *string, const char *format, va_list ap);
typedef int (*vfprintf_t) (FILE *stream, const char *format, va_list ap);
typedef int (*vsnprintf_t) (char *string, size_t size, const char *format, va_list ap);
typedef void (*vsyslog_t) (int priority, const char *format, va_list ap);
typedef void (*vwarn_t) (const char *format, va_list ap);
typedef void (*vwarnx_t) (const char *format, va_list ap);
typedef void (*verr_t) (int status, const char *format, va_list ap);
typedef void (*verrx_t) (int status, const char *format, va_list ap);

/*
 * Avoid locale-dependent isdigit()
 */
#define myisdigit(x) (strchr("0123456789",(x)) != NULL)

/*
 * Flag to disable checking for our internal routines so we can still
 * log even though we put a wrapper around syslog.
 */
static int disable_check = 0;

/*
 * Fatal error handler; log to syslog and destroy the current process.
 * (Does not return)
 *
 * SIGKILL was chosen because it cannot be caught, ignored and terminates
 * the process.
 */
static void fatal (const char *fmt, ...)
{
	va_list ap;
	char *newfmt, *pre = "libformat: ";

	/*
	 * Prefix the format string with the name of our library so messages
	 * we generate can be distinguished from other messages generated by
	 * the programs we protect.
	 */
	newfmt = (char *) malloc (strlen (fmt) + strlen (pre) + 1);
	if (newfmt == NULL) {
		/*
		 * We _really_ don't want to fail here. Log a warning message
		 * and use the old format string.
		 */
		newfmt = (char *) fmt;
	} else {
		strcpy (newfmt, pre);
		strcat (newfmt, fmt);
	}

	disable_check = 1;
	va_start (ap, fmt);
	vsyslog (LOG_ERR, newfmt, ap);
	va_end (ap);

	syslog (LOG_ERR|LOG_PID, "libformat: sending current process the "
		"KILL signal");

	(void) raise (SIGKILL);
	/*
	 * Let's hope raise() does not fail.
	 */
}

/*
 * Get a pointer to the real function named `funcname' from libc.
 */
static void *get_libc_func (const char *funcname)
{
	void *func;
	char *error;

	func = dlsym (RTLD_NEXT, funcname);
	if ((error = dlerror()) != NULL) {
		fatal ("can't find function `%s' in libc: %s",
				funcname, error);
	}

	return func;
}

/*
 * Check whether the format string contains %n conversions, and if so, make
 * sure that the string is not in a writeable segment. %n characters are a
 * sign that someone might be trying to attack the program.
 */
static void check_format (const char *format)
{
	FILE *fp;
	char line[512];
	const char *addr;
	char *sp;
	size_t numargs, maxargs, i;
	int *argtypes;
	int found_n = 0;

	/*
	 * Don't check if checking is disabled or if the format string is
	 * NULL.
	 */
	if ((disable_check != 0) || (format == NULL))
		return;

	/*
	 * Check for bad conversions (%n). We now use a glibc-specific
	 * function to parse printf format strings because it's too tricky
	 * to do properly ourselves.
	 */
	maxargs = (strlen (format) / 2);
	argtypes = (int *) alloca (maxargs * sizeof (int));
	if (argtypes == NULL)
		fatal ("can't allocate memory for parse_printf_format");
	numargs = parse_printf_format (format, maxargs, argtypes);
	
	if (numargs == 0)
		/*
		 * No conversions, no need to check anything.
		 */
		return;
	
	if (numargs > maxargs)
		/*
		 * Just in case - this really should not ever happen.
		 */
		fatal ("too many conversions in format string");
	
	for (i = 0; i < numargs; i++)
		if (((argtypes[i] &~ PA_FLAG_MASK) == PA_INT) && (argtypes[i] & PA_FLAG_PTR)) found_n = 1;

	if (found_n == 0)
		/*
		 * This format string has no bad conversions, let it through.
		 */
		return;

	if ((fp = fopen ("/proc/self/maps", "r")) == NULL)
		fatal ("/proc/self/maps: %s", strerror (errno));

        /* 
	 * Find the segment that the format string is in and make sure it's
	 * not writeable.
	 *
	 * Lines in /proc/self/maps look like this right now:
	 * start-end [-r][-w][-x][-p] flags maj:min inode pathname 
	 *
	 * This is derived from code posted by Olaf Kirch to the
	 * security-audit mailing list. It has been hacked beyond recognition
	 * by myself.
         */
        while (fgets(line, (int) sizeof(line), fp) != NULL) {
		if (line[strlen(line) - 1] != '\n')
			fatal ("/proc/self/maps: line too long");
		addr = (char *) strtoul(line, &sp, 16);
		if (addr > format) continue;
		if (*sp++ != '-')
			fatal ("/proc/self/maps: incorrect format");

		addr = (char *) strtoul(sp, &sp, 16);
		if (addr <= format) continue;
		if (*sp++ != ' ')
			fatal ("/proc/self/maps: incorrect format");

		/*
		 * We've found the segment that contains the format string,
		 * check whether it's writeable. If so, fatal() will terminate
		 * the current process and let the administrator know what
		 * happened.
		 */
		if (sp[1] == 'w')
			fatal ("non-constant format string `%.50s'%s used", 
					format,
					(strlen (format) > 50) ? "..." : "");

		(void) fclose (fp);
		return;
	}

	fatal ("format string is not in any of current process' segments");
}

/*
 * Wrappers around printf family of functions:
 * printf, vprintf, sprintf, vsprintf, fprintf, vfprintf, snprintf, vsnprintf
 */

int vprintf (const char *format, va_list ap)
{
	vprintf_t real_vprintf;
	
	real_vprintf = (vprintf_t) get_libc_func ("vprintf");
	check_format (format);
	
	return real_vprintf (format, ap);
}

int printf (const char *format, ...)
{
	va_list ap;
	int retval;

	va_start (ap, format);
	retval = vprintf (format, ap); /* our wrapper */
	va_end (ap);

	return retval;
}

int vsprintf (char *string, const char *format, va_list ap)
{
	vsprintf_t real_vsprintf;

	real_vsprintf = (vsprintf_t) get_libc_func ("vsprintf");
	check_format (format);

	return real_vsprintf (string, format, ap);
}

int sprintf (char *string, const char *format, ...)
{
	va_list ap;
	int retval;

	va_start (ap, format);
	retval = vsprintf (string, format, ap); /* our wrapper */
	va_end (ap);

	return retval;
}

int vsnprintf (char *string, size_t size, const char *format, va_list ap)
{
	vsnprintf_t real_vsnprintf;
	
	real_vsnprintf = (vsnprintf_t) get_libc_func ("vsnprintf");
	check_format (format);
	
	return real_vsnprintf (string, size, format, ap);
}

int snprintf (char *string, size_t size, const char *format, ...)
{
	va_list ap;
	int retval;

	va_start (ap, format);
	retval = vsnprintf (string, size, format, ap); /* our wrapper */
	va_end (ap);

	return retval;
}

int vfprintf (FILE *stream, const char *format, va_list ap)
{
	vfprintf_t real_vfprintf;

	real_vfprintf = (vfprintf_t) get_libc_func ("vfprintf");
	check_format (format);

	return real_vfprintf (stream, format, ap);
}

int fprintf (FILE *stream, const char *format, ...)
{
	va_list ap;
	int retval;

	va_start (ap, format);
	retval = vfprintf (stream, format, ap); /* our wrapper */
	va_end (ap);

	return retval;
}


/*
 * Wrappers around syslog family of functions:
 * syslog vsyslog
 */

void vsyslog (int priority, const char *format, va_list ap)
{
	vsyslog_t real_vsyslog;

	real_vsyslog = (vsyslog_t) get_libc_func ("vsyslog");
	check_format (format);

	real_vsyslog (priority, format, ap);
}

void syslog (int priority, const char *format, ...)
{
	va_list ap;

	va_start (ap, format);
	vsyslog (priority, format, ap);
	va_end (ap);
}

/*
 * Wrappers around warn/err family of functions:
 * warn vwarn warnx vwarnx err verr errx verrx
 */

void vwarn (const char *format, va_list ap)
{
	vwarn_t real_vwarn;

	real_vwarn = (vwarn_t) get_libc_func ("vwarn");
	check_format (format);

	real_vwarn (format, ap);
}

void warn (const char *format, ...)
{
	va_list ap;

	va_start (ap, format);
	vwarn (format, ap);
	va_end (ap);
}

void vwarnx (const char *format, va_list ap)
{
	vwarnx_t real_vwarnx;

	real_vwarnx = (vwarnx_t) get_libc_func ("vwarnx");
	check_format (format);

	real_vwarnx (format, ap);
}

void warnx (const char *format, ...)
{
	va_list ap;

	va_start (ap, format);
	vwarnx (format, ap);
	va_end (ap);
}

void verr (int status, const char *format, va_list ap)
{
	verr_t real_verr;

	real_verr = (verr_t) get_libc_func ("verr");
	check_format (format);

	real_verr (status, format, ap);
}

void err (int status, const char *format, ...)
{
	va_list ap;

	va_start (ap, format);
	verr (status, format, ap);
	va_end (ap);
}

void verrx (int status, const char *format, va_list ap)
{
	verrx_t real_verrx;

	real_verrx = (verrx_t) get_libc_func ("verrx");
	check_format (format);

	real_verrx (status, format, ap);
}

void errx (int status, const char *format, ...)
{
	va_list ap;

	va_start (ap, format);
	verrx (status, format, ap);
	va_end (ap);
}

/*
 * TODO: Wrap the following functions:
 * <curses.h>: mvprintw mvwprintw printw tparm wprintw
 * <stdio.h>: dprintf?
 *
 * Stop duplicating so much code!
 */
